#include "entities/bullet.hpp"
#include "entities/enemy.hpp"
#include "entities/tower.hpp"
#include <algorithm>
#include <cstdlib>
#include <gint/cpu.h>
#include <gint/display.h>
#include <gint/keyboard.h>
#include <gint/timer.h>
#include <vector>

#define FRAME_RATE 30

bool ExitToOS = false;

Tower *MyTower;
std::vector<Enemy *> MyEnemies;
std::vector<Bullet *> MyBullets;

volatile int frame_tick = 1;
int timer_id;

void SetTimer(void) {

  timer_id = timer_configure(TIMER_ANY, 1000000 / FRAME_RATE,
                             GINT_CALL_SET(&frame_tick));
  if (timer_id >= 0)
    timer_start(timer_id);
}

void CloseTimer(void) {
  if (timer_id >= 0)
    timer_stop(timer_id);
}

void Render(void) {
  dclear(C_WHITE);
  MyTower->Render();

  for (auto &e : MyEnemies)
    e->Render();

  for (auto &b : MyBullets)
    b->Render();

  dprint(1, 1, C_BLACK, "Nb Enemies = %i", MyEnemies.size());
  dprint(1, 11, C_BLACK, "Nb Bullets = %i", MyBullets.size());

  dupdate();
}

void Update(void) {
  // Update player's tower
  MyTower->Update();

  // sort vector of Enemies by distance from the tower
  sort(MyEnemies.begin(), MyEnemies.end(),
       [](const Enemy *lhs, const Enemy *rhs) {
         return lhs->DistanceToTower < rhs->DistanceToTower;
       });

  // Update Bullets
  for (auto &b : MyBullets) {
    b->Update();
    // Check for the current Bullet if hits an Enemy
    for (auto &e : MyEnemies)
      if (e->TestImpact(b)) {
        b->HitDone();
      }
  }

  // Update Enemies
  for (auto &e : MyEnemies)
    e->Update();

  // Remove Enemies killed
  MyEnemies.erase(
      std::remove_if(MyEnemies.begin(), MyEnemies.end(),
                     [](const Enemy *lhs) { return lhs->ToBeDeleted; }),
      MyEnemies.end());

  // remove Bullets that hit an Enemy
  MyBullets.erase(
      std::remove_if(MyBullets.begin(), MyBullets.end(),
                     [](const Bullet *lhs) { return lhs->ToBeDeleted; }),
      MyBullets.end());
}

void GetInputs(void) {
  cleareventflips();
  clearevents();
  if (keypressed(KEY_EXIT))
    ExitToOS = true;
  if (keypressed(KEY_F1)) {
    MyEnemies[0]->ToBeDeleted = true;
  }
  if (keypressed(KEY_F2)) {
    Bullet *b = new Bullet(MyEnemies[0]->EnemyX, MyEnemies[0]->EnemyY);
    MyBullets.push_back(b);
  }
}

int main(void) {

  srand(0);

  SetTimer();

  MyTower = new Tower();

  for (int i = 0; i < 25; i++) {
    Enemy *E = new Enemy();
    MyEnemies.push_back(E);
  }

  while (!ExitToOS) {
    while (!frame_tick)
      sleep();
    GetInputs();
    Update();
    Render();
  }

  CloseTimer();

  delete MyTower;

  for (auto &e : MyEnemies)
    delete (e);
  MyEnemies.clear();

  return 1;
}
